@@ -1,8 +1,10 @@
 import json
+import os.path as osp
 import queue
 import threading
 import tkinter as tk
-from tkinter import ttk, filedialog
+import types
+from tkinter import ttk, filedialog, scrolledtext as tktext
 from tkinter import messagebox as tkmsgbox
 # 3rd party
 import kkpyutil as util
@@ -11,32 +13,48 @@ import kkpyutil as util
 class Globals:
     root = None
     progressQueue = queue.Queue()
+    taskStopEvent = threading.Event()
 
 
-def _validate_int(user_input, new_value, widget_name):
-    return _validate_number(user_input, new_value, widget_name, int)
+def _validate_int(value_after_input, user_input, widget_name):
+    return _validate_number(value_after_input, user_input, widget_name, int)
 
 
-def _validate_float(user_input, new_value, widget_name):
-    return _validate_number(user_input, new_value, widget_name, float)
+def _validate_float(value_after_input, user_input, widget_name):
+    return _validate_number(value_after_input, user_input, widget_name, float)
 
 
-def _validate_number(user_input, new_value, widget_name, data_type):
-    # disallow anything but numbers in the input
-    is_digit = new_value == '' or new_value.isdigit()
-    if not is_digit:
+def _validate_number(value_after_input, user_input, widget_name, data_type):
+    """
+    - ttk.spinbox does not support paste over selection; pasted content always prepends
+    - so pasting always fails number validation in spinbox
+    """
+    # allow '-6.' and append 0 to fix format
+    if truncated_float := data_type == float and value_after_input.endswith('.') and util.is_float_text(value_after_input):
+        value_after_input += '0'
+        return True
+    # '-' is not a valid number and will fail here, beware erase digits in a row with backspace
+    if not util.is_number_text(value_after_input) or value_after_input == '':
         Globals.root.bell()
         return False
-    minval = data_type(Globals.root.nametowidget(widget_name).config('from')[4])
-    maxval = data_type(Globals.root.nametowidget(widget_name).config('to')[4])
-    if not (minval <= data_type(user_input) <= maxval):
+    raw_min = Globals.root.nametowidget(widget_name).config('from')[4]
+    raw_max = Globals.root.nametowidget(widget_name).config('to')[4]
+    minval = raw_min if raw_min in (float('-inf'), float('inf')) else data_type(raw_min)
+    maxval = raw_max if raw_max in (float('-inf'), float('inf')) else data_type(raw_max)
+    try:
+        if not (minval <= data_type(value_after_input) <= maxval):
+            Globals.root.bell()
+            util.alert(f"""New value {value_after_input} would fall outside of range: [{minval}, {maxval}]
+Change skipped""", 'ERROR')
+            return False
+    except ValueError as e:
         Globals.root.bell()
         return False
     return True
 
 
 class Root(tk.Tk):
-    def __init__(self, title, size=(800, 600), *args, **kwargs):
+    def __init__(self, title, size=(800, 600), icon=None, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.title(title)
         screen_size = (self.winfo_screenwidth(), self.winfo_screenheight())
@@ -48,24 +66,40 @@ class Root(tk.Tk):
         )
         self.validateIntCmd = (self.register(_validate_int), '%P', '%S', '%W')
         self.validateFloatCmd = (self.register(_validate_float), '%P', '%S', '%W')
+        if icon:
+            self.iconphoto(True, tk.PhotoImage(file=icon))
+        self.controller = None
+        # used by on_during_deactivate only for the shutdown sequence only
+        self.isActive = False
         self._auto_focus()
 
-    def bind_events(self, controller):
+    def set_controller(self, controller):
         """
-        - controller interface, must implement:
+        - controller is used in many non-contiguous calls in the init sequence
+        - so for DRY, it needs to be a member
+        - it's created after root because views that it knows must use root as their parent,
+        - so we set it using a setter whenever there is a chance
+        """
+        self.controller = controller
+
+    def bind_events(self):
+        """
+        - controller interface must implement:
           - ENTER key event: default action
           - ESC key event: cancel/negate default action
           - Window X button event: quit
         - refer to Inter-Client Communication Conventions Manual ICCCM for possible window events
         """
-        self.bind("<Return>", controller.submit)
-        self.bind("<Escape>", lambda event: controller.cancel(event))
+        self.bind("<Return>", self.controller.on_submit)
+        self.bind("<Escape>", lambda event: self.controller.on_cancel(event))
         # Expose: called even when slider is dragged, so we don't use it
-        # Map: triggered when windows is visible
-        self.bind('<Map>', lambda event: controller.init(event))
-        self.bind('<Destroy>', lambda event: controller.term(event))
+        # Map: triggered when windows are visible, called every frame
+        # Destroy: triggered when windows are closed, called every frame
+        self.bind('<Map>', lambda event: self.on_during_activate(event))
+        self.bind('<Destroy>', lambda event: self.on_during_deactivate(event))
+        # startup event: init(), must be called by client
         # bind X button to quit the program
-        self.protocol('WM_DELETE_WINDOW', controller.quit)
+        self.protocol('WM_DELETE_WINDOW', self.controller.on_quit)
 
     def _auto_focus(self):
         def _unpin_root(event):
@@ -79,21 +113,50 @@ class Root(tk.Tk):
         self.focus_force()
         self.bind('<FocusIn>', _unpin_root)
 
+    def mainloop(self, n: int = 0):
+        """
+        - prepend custom pre-startup event
+        - this solves the problem where <Map> event, being a per-frame activation event, gets called many times instead of just once, which is redundant for startup logic
+        """
+        self.controller.update_model()
+        self.after(0, self.controller.on_startup)
+        super().mainloop()
+
+    def on_during_activate(self, event):
+        """
+        - called every frame during the root window display process, i.e., from background to foreground
+        - but since app-level startup tasks only need to be done once, we bootstrap the frame-behavior for controller to perform app-level tasks
+        """
+        if self.isActive:
+            return
+        self.controller.on_activate(event)
+        self.isActive = True
+
+    def on_during_deactivate(self, event):
+        """
+        - similar to on_during_activate(), but for shutdown
+        """
+        if not self.isActive:
+            return
+        self.controller.on_deactivate(event)
+        self.isActive = False
+
 
 class Prompt:
     """
     - must use within tkinter mainloop
-    - otherwise will hang upon confirmation
+    - otherwise the app will freeze upon confirmation
     """
 
-    def __init__(self, logger=None):
+    def __init__(self, master=Globals.root, logger=None):
+        self.master = master
         self.logger = logger or util.glogger
 
     def info(self, msg, confirm=True):
         """Prompt with info."""
         self.logger.info(msg)
         if confirm:
-            tkmsgbox.showinfo('Info', msg, icon='info')
+            tkmsgbox.showinfo('Info', msg, icon='info', parent=self.master)
 
     def warning(self, detail, advice, question='Continue?', confirm=True):
         """
@@ -111,12 +174,12 @@ Advice:
         self.logger.warning(msg)
         if not confirm:
             return True
-        return tkmsgbox.askyesno('Warning', msg, icon='warning')
+        return tkmsgbox.askyesno('Warning', msg, icon='warning', parent=self.master)
 
     def error(self, errclass, detail, advice, confirm=True):
         """
         - for problems with significant impact
-        - program will crash immediately
+        - the program will crash immediately if not to confirm
         """
         msg = f"""\
 Detail:
@@ -128,7 +191,7 @@ Advice:
 Will crash"""
         self.logger.error(msg)
         if confirm:
-            tkmsgbox.showerror('Error', msg, icon='error')
+            tkmsgbox.showerror('Error', msg, icon='error', parent=self.master)
         raise errclass(msg)
 
 
@@ -209,7 +272,7 @@ class Form(ttk.PanedWindow):
     - layout: page-based navigation
     - filter: locate form entries by searching for title keywords
     - structure: Form > Page > Entry
-    - instantiation: Form > Page (slave to form pane) > Entry (slave to page)
+    - instantiation: Form > Page (slaved to form pane) > Entry (slaved to page)
     """
 
     def __init__(self, master, page_titles: list[str], **kwargs):
@@ -230,12 +293,11 @@ class Form(ttk.PanedWindow):
         self.tree.pack(side="left", fill="both", expand=True)
         self.tree.bind("<<TreeviewSelect>>", self.update_entries)
         # Right panel: entries in page
-        # self.entryPane = ttk.Frame(self)
         self.entryPane = ScrollFrame(self)
         # build form with navbar and page frame
         self.add(self.navPane, weight=0)
         self.add(self.entryPane, weight=1)
-        self.pages = {title: Page(self.entryPane.frame, title) for title in page_titles}
+        self.pages = {title.lower(): Page(self.entryPane.frame, title.title()) for title in page_titles}
         self.init()
         self.layout()
 
@@ -245,7 +307,7 @@ class Form(ttk.PanedWindow):
     def init(self):
         # Populate tree with page titles
         for title, pg in self.pages.items():
-            self.tree.insert("", "end", text=title)
+            self.tree.insert("", "end", text=title.title())
         # select first page
         self.tree.selection_set(self.tree.get_children()[0])
         self.update_entries(None)
@@ -262,7 +324,7 @@ class Form(ttk.PanedWindow):
         for pg in self.pages.values():
             pg.pack_forget()
         # After hiding, update the right pane to ensure correct display
-        self.pages[selected_title].layout() if selected_title else list(self.pages.values())[0].layout()
+        self.pages[selected_title.lower()].layout() if selected_title else list(self.pages.values())[0].layout()
         self.entryPane.update()
 
     def _on_clear_search(self, event):
@@ -302,33 +364,37 @@ class Entry(ttk.Frame):
     - used as user input, similar to CLI arguments
     - widget must belong to a group
     - groups form a tree to avoid overloading parameter panes
-    - groups also improves SNR by prioritizing frequently-tweaked parameters
+    - groups also improve SNR by prioritizing frequently-tweaked parameters
     - page is responsible for lay out entries
     """
 
-    def __init__(self, master: Page, text, widget_constructor, default, doc, **widget_kwargs):
+    def __init__(self, master: Page, key, text, widget_constructor, default, doc, configable=True, **widget_kwargs):
         super().__init__(master)
         assert isinstance(self.master, Page)
         self.master.add([self])
+        self.key = key
         self.text = text
         self.default = default
+        self.isConfig = configable
         # model-binding
         self.data = None
         # title
         self.label = ttk.Label(self, text=self.text, cursor='hand2')
-        self.label.grid(row=0, column=0, sticky='w')
+        self.label.pack(expand=True, padx=5, pady=2, anchor="w")
         self.label.bind("<Double-Button-1>", lambda e: tkmsgbox.showinfo("Help", doc))
         # field
         self.field = widget_constructor(self, **widget_kwargs)
         self.columnconfigure(0, weight=1)
-        self.field.grid(row=1, column=0, sticky='ew', padx=5, pady=5)
+        self.field.pack(expand=True, padx=5, pady=2, anchor="w")
         # context menu
         self.contextMenu = tk.Menu(self, tearoff=0)
-        # use context menu instead of direct clicking to avoid accidental reset
+        # use a context menu instead of direct clicking to avoid accidental reset
         self.contextMenu.add_command(label="Reset", command=self.reset)
         # maximize context-menu hitbox
         self.field.bind("<Button-2>", self.show_context_menu)
         self.label.bind("<Button-2>", self.show_context_menu)
+        # getting out of focus so that key strokes will not be intercepted by the entry
+        self.field.bind("<Escape>", lambda event: Globals.root.focus_set())
 
     def _init_data(self, var_cls):
         return var_cls(master=self, name=self.text, value=self.default)
@@ -343,7 +409,7 @@ class Entry(ttk.Frame):
         self.data.set(value)
 
     def layout(self):
-        self.pack(fill="both", expand=True, padx=5, pady=15, anchor="w")
+        self.pack(fill="both", expand=True, padx=5, pady=10, anchor="w")
 
     def show_context_menu(self, event):
         try:
@@ -353,11 +419,11 @@ class Entry(ttk.Frame):
 
     def set_tracer(self, handler):
         """
-        - handler: callback(name, var, index, mode)
+        - handler: callback (name, var, index, mode)
           - name: name of the variable
-          - var: tk.variable object
+          - var: tk.Variable object
           - index: index of the variable
-          - mode: 'read'(triggered when var is read), 'write'(triggered when var is written), 'unset'
+          - mode: 'read' (triggered when var is read), 'write'(triggered when var is written), 'unset'
         """
         self.data.trace_add('write', callback=lambda name, index, mode, var=self.data: handler(name, var, index, mode))
 
@@ -370,8 +436,16 @@ class FormMenu(tk.Menu):
         self.fileMenu = tk.Menu(self, tearoff=False)
         self.fileMenu.add_command(label="Load Preset ...", command=self.on_load_preset)
         self.fileMenu.add_command(label="Save Preset ...", command=self.on_save_preset)
-        self.fileMenu.add_command(label="Exit", command=self.on_quit)
+        self.fileMenu.add_command(label="Quit", command=self.on_quit, accelerator="Ctrl+Q")
+        self.master.bind("<Control-q>", lambda event: self.on_quit())
+        self.master.bind("<Control-Q>", lambda event: self.on_quit())
+        self.helpMenu = tk.Menu(self, tearoff=False)
+        self.helpMenu.add_command(label="Help", command=self.on_open_help, accelerator="F1")
+        self.helpMenu.add_command(label="Open Log", command=self.on_open_log)
+        self.helpMenu.add_command(label="Report A Problem", command=self.on_report_issue)
+        self.master.bind("<F1>", lambda event: self.on_open_help())
         self.add_cascade(label="File", menu=self.fileMenu)
+        self.add_cascade(label="Help", menu=self.helpMenu)
         self.controller = controller
 
     def on_load_preset(self):
@@ -381,108 +455,229 @@ class FormMenu(tk.Menu):
             ("Preset Files", "*.json"),
         ])
         if preset:
-            self.controller.load(preset)
+            self.controller.load_preset(preset)
 
     def on_save_preset(self):
         preset = filedialog.asksaveasfilename(title="Save Preset", filetypes=[
             ("Preset Files", "*.preset.json"),
         ])
         if preset:
-            self.controller.save(preset)
+            self.controller.save_preset(preset)
+
+    def on_open_help(self):
+        self.controller.on_open_help()
+
+    def on_open_log(self):
+        self.controller.on_open_log()
+
+    def on_report_issue(self):
+        self.controller.on_report_issue()
 
     def on_quit(self):
-        self.controller.quit(None)
+        self.controller.on_quit(None)
 
 
 class FormController:
     """
     - observe all entries and update model
+    - both form and realtime apps can use this class
+    - form apps can use submit() to update model
+    - realtime apps can set arg-tracers
+    - model and app-config share the same keys
+    - backend task works in task thread
+    - progressbar and task synchronize via threading.Event
     """
 
-    def __init__(self, fm=None, model=None):
-        self.form = fm
+    def __init__(self, form=None, model=None):
+        self.form = form
         self.model = model
+        self.set_progress = lambda title, progress, description: Globals.progressQueue.put((title, progress, description))
+        self.taskThread = None
+        self.taskStopEvent = Globals.taskStopEvent
 
-    def update(self):
-        self.model = {pg.get_title(): {entry.text: entry.get_data() for entry in pg.winfo_children()} for pg in self.form.pages.values()}
+    def update_model(self):
+        config_by_page = {
+            pg.get_title(): {entry.key: entry.get_data() for entry in pg.winfo_children()}
+            for title, pg in self.form.pages.items()
+        }
+        self.model = {k: v for entries in config_by_page.values() for k, v in entries.items()}
 
-    def load(self, preset):
+    def update_view(self):
+        self.load_preset(self.model)
+
+    def load_preset(self, preset):
         """
         - model includes input and config
         - input is runtime data that changes with each run
         - only config will be saved/loaded as preset
         """
-        config = util.load_json(preset)
+        config = util.load_json(preset) if isinstance(preset, str) else preset
         for title, page in self.form.pages.items():
             for entry in page.winfo_children():
                 try:
-                    entry.set_data(config[title][entry.text])
-                except KeyError:
-                    pass
+                    entry.set_data(config[entry.key])
+                except KeyError as e:
+                    util.glogger.error(e)
 
-    def save(self, preset):
+    def save_preset(self, preset):
         """
         - only config is saved
         - input always belongs to group "input"
+        - in app-config, if user specifies title, then the title is used with presets (titlecase) instead of the original key (lowercase)
         """
-        self.update()
-        config = {pg.get_title(): {entry.text: entry.get_data() for entry in pg.winfo_children()} for title, pg in self.form.pages.items() if title != "input"}
+        config_by_page = {
+            pg.get_title(): {entry.key: entry.get_data() for entry in pg.winfo_children() if entry.isConfig}
+            for title, pg in self.form.pages.items()
+        }
+        config = {k: v for entries in config_by_page.values() for k, v in entries.items()}
         util.save_json(preset, config)
 
-    def reset(self):
+    def is_scheduled_to_stop(self):
+        return self.taskStopEvent.is_set()
+
+    def start_progress(self):
+        self.set_progress('/start', 0, 'Processing ...')
+
+    def stop_progress(self):
+        """
+        - progressbar will stop where it is at the moment
+        """
+        self.set_progress('/stop', 100, 'Stopped')
+
+    def get_latest_model(self):
+        """
+        - for easy consumption of client objects as arg
+        """
+        self.update_model()
+        return types.SimpleNamespace(**self.model)
+
+    def wait_for_task(self, wait_ms=100):
+        if self.taskThread.is_alive():
+            # Schedule this method to be called again after wait_ms milliseconds
+            self.form.after(wait_ms, self.wait_for_task)
+
+    #
+    # callbacks
+    #
+    def on_open_help(self):
+        """
+        - open help doc, e.g., webpage, local file
+        - subclass this for your own 
+        """
+        prompt = Prompt()
+        prompt.info('Help not implemented yet; implement it in controller subclasses', confirm=True)
+
+    def on_open_log(self):
+        """
+        - open log or app session data is hard to generalize
+        - subclass this to use app-level logging scheme
+        - e.g., opening a log file using the default browser
+        - e.g., opening a folder containing the entire diagnostics
+        """
+        prompt = Prompt()
+        prompt.info('Logging not implemented yet; implement it in controller subclasses', confirm=True)
+
+    def on_report_bug(self):
+        """
+        - report bug to the developer
+        - subclass this
+        """
+        prompt = Prompt()
+        prompt.info('Bug reporting not implemented yet; implement it in controller subclasses', confirm=True)
+
+    def on_reset(self):
+        """
+        - reset all form fields to default
+        - usually can be used as is, no need to override
+        """
         for pg in self.form.pages.values():
             for entry in pg.winfo_children():
                 entry.reset()
 
-    def submit(self, event=None):
+    def on_submit(self, event=None):
         """
-        - subclass this to implement custom logic
+        - main action to launch the background task
+        - usually can be used as is, no need to override
         """
-        self.update()
+        if self.taskThread and self.taskThread.is_alive():
+            return
+        self.update_model()
+        self.taskStopEvent.clear()
         # lambda wrapper ensures "self" is captured by threading as a context
         # otherwise ui thread still blocks
-        threading.Thread(target=lambda: self.run_background(), daemon=True).start()
+        self.taskThread = threading.Thread(target=lambda: self.run_task(), daemon=True)
+        self.taskThread.start()
 
-    def run_background(self):
+    def run_task(self):
         """
         - override this in app
-        - run in background thread to avoid blocking UI
+        - run in background thread to unblock UI
         """
         raise NotImplementedError('subclass this!')
 
-    def cancel(self, event=None):
+    def on_cancel(self, event=None):
         """
-        - for form-filling case, we quit window on cancelling
-        - override this in app
+        - cancelling a running background task
         """
-        self.quit()
+        if self.taskThread and self.taskThread.is_alive():
+            self.taskStopEvent.set()
+            self.wait_for_task()
 
-    def quit(self, event=None):
+    def on_quit(self, event=None):
         """
-        - CAUTION:
-          - for sharing binding between menu, x-button, and other ui for quitting
-          - although usually we avoid physical ops in controller
-        - override term() in app
+        CAUTION:
+        - usually we avoid direct view-ops in controller
+        - but here it is necessary for sharing binding between menu, x-button, and other quitting devi
         """
-        self.term()
+        if not self.on_shutdown():
+            # user cancelled
+            return
         self.form.master.quit()
 
-    def init(self, event=None):
+    def on_startup(self):
         """
-        - binding of <Expose> event as logical initialization
-        - called after mainloop() started, i.e., when window is partially visible
-        - controller can start retrieving entries
-        - override this in app
+        - called just before showing root window (<Map>, on_activate()), after all fields are initialized
+        - so that fields can be used here for the first time
         """
         pass
 
-    def term(self, event=None):
+    def on_shutdown(self) -> bool:
+        """
+        - called just before quitting
+        - safe-schedules shutdown with prompt and early-outs if user cancels
+        - subclass this for post-ops
+        """
+        if not self.taskThread or not self.taskThread.is_alive():
+            # task not running, safe to continue to quit
+            self.taskStopEvent.set()  # progressbar needs to be stopped
+            return True
+        prompt = Prompt()
+        # Make default behavior a safe bet
+        if prompt.warning('Quitting a running task may cause damage. Click Yes to wait for it to finish, or No to force-quit', 'Wait for it to finish.', question='Keep waiting?', confirm=True):
+            # user cancelled
+            return False
+        self.taskStopEvent.set()  # progressbar needs to be stopped
+        # task should have received stop event, let's wait for it to end
+        # it may choose a safe-quit path, but maybe not (damage)
+        self.wait_for_task()
+        return True
+
+    def on_activate(self, event=None):
+        """
+        - binding of <Map> event as logical initialization
+        - called once when the root window displays, i.e., from background to foregrounded
+        """
+        pass
+
+    def on_deactivate(self, event=None):
         """
         - binding of <Destroy> event as logical termination
         - called AFTER triggering WM_DELETE_WINDOW
-        - override this in app
+        - called once when the root window disappears, from foreground to background
+        - on macOS: called on Cmd+Q key-combo, which quits python launcher and bypasses WM_DELETE_WINDOW
         """
-        pass
+        if util.PLATFORM == 'Darwin':
+            self.on_quit()
 
 
 class FormActionBar(ttk.Frame):
@@ -492,11 +687,11 @@ class FormActionBar(ttk.Frame):
         self.controller = controller
         # occupy the entire width
         # new buttons will be added to the right
-        self.resetBtn = ttk.Button(self, text="Reset", command=self.on_reset_entries)
+        self.resetBtn = ttk.Button(self, text="Reset", command=self.on_reset)
         self.separator = ttk.Separator(self, orient="horizontal")
         # Create Cancel and Submit buttons
-        self.cancelBtn = ttk.Button(self, text="Cancel", command=self.on_cancel)
-        self.submitBtn = ttk.Button(self, text="Submit", command=self.on_submit, cursor='hand2')
+        self.cancelBtn = ttk.Button(self, text="Stop", command=self.on_cancel)
+        self.submitBtn = ttk.Button(self, text="Start", command=self.on_submit, cursor='hand2')
         # layout: keep the order
         self.separator.pack(fill="x")
         # left-most must pack after separator to avoid occluding the border
@@ -508,32 +703,20 @@ class FormActionBar(ttk.Frame):
     def layout(self):
         self.pack(side="bottom", fill="x")
 
-    def submit(self, event=None):
-        """for debugging only"""
-        self.controller.update()
-        formatted_data = json.dumps(self.controller.model, indent=4)
-        tkmsgbox.showinfo("Submitted Data", formatted_data)
-
-    def on_reset_entries(self, event=None):
-        self.controller.reset()
+    def on_reset(self, event=None):
+        self.controller.on_reset()
 
     def on_cancel(self, event=None):
-        self.controller.cancel()
+        self.controller.on_cancel()
 
     def on_submit(self, event=None):
-        self.controller.submit()
-
-
-class OnOffActionBar(FormActionBar):
-    def __init__(self, master, controller, *args, **kwargs):
-        super().__init__(master, controller, *args, **kwargs)
-        self.submitBtn.configure(text='Start')
-        self.cancelBtn.configure(text='Stop')
+        self.controller.on_submit()
 
 
 class WaitBar(ttk.Frame):
     """
     - app must run in worker thread to avoid blocking UI
+    - use /start, /stop, /processing to mark start/end/progress
     - when using subprocess to run a blackbox task, use indeterminate mode cuz there is no way to pass progress back
     - protocol: tuple(stage, progress, description), where stage is program instruction, description is for display
     - TODO: use IPC for cross-language open-source tasks
@@ -545,6 +728,7 @@ class WaitBar(ttk.Frame):
         self.stage = tk.StringVar(name='stage', value='')
         self.bar = ttk.Progressbar(self, orient="horizontal", mode="indeterminate")
         self.label = ttk.Label(self.bar, textvariable=self.stage, text='...', foreground='white', background='black')
+        self.taskStopEvent = Globals.taskStopEvent
         self.layout()
 
     def layout(self):
@@ -555,21 +739,31 @@ class WaitBar(ttk.Frame):
         self.label.place(relx=0.5, rely=0.5, anchor='center')
         self.pack(side='bottom', fill='both', expand=False)
 
-    def poll(self, wait_ms=50):
+    def poll(self, wait_ms=100):
         """
         - app pushes special messages to mark progress start/stop
         """
         while self.queue.qsize():
+            if self._is_scheduled_to_stop():
+                return
             msg = self.queue.get(0)
             cmd = msg[0]
+            # CAUTION:
+            # - .start() and .stop() are used for indeterminate progress bars only
+            # - use .set() with determinate bars
             if cmd == '/start':
                 self.bar.start()
+                self.master.update_idletasks()
             elif cmd == '/stop':
                 self.bar.stop()
+                self.master.update_idletasks()
             else:
                 raise NotImplementedError(f'Unexpected progress instruction: {cmd}')
         self.after(wait_ms, self.poll)
 
+    def _is_scheduled_to_stop(self):
+        return self.taskStopEvent.is_set()
+
 
 class ProgressBar(WaitBar):
     def __init__(self, master, progress_queue, *args, **kwargs):
@@ -578,18 +772,14 @@ class ProgressBar(WaitBar):
         self.bar.configure(variable=self.progress, mode='determinate')
         self.layout()
 
-    def poll(self, wait_ms=50):
+    def poll(self, wait_ms=100):
         """
         - Periodically check for messages from worker thread.
         """
         while self.queue.qsize():
             try:
                 cmd, value, text = self.queue.get_nowait()
-                if cmd == '/start':
-                    self.bar.start()
-                elif cmd == '/stop':
-                    self.bar.stop()
-                elif cmd == '/processing':
+                if cmd == '/processing':
                     self.progress.set(value)
                 self.stage.set(text)  # Update the label text
             except queue.Empty:
@@ -600,11 +790,11 @@ class ProgressBar(WaitBar):
 class IntEntry(Entry):
     """
     - show slider for finite numbers only
-    - ttk.Scale is intended for ratio only; the handle does not move for negative numbers
+    - ttk.Scale is intended for a ratio only; the handle does not move for negative numbers
     - must bind a separate variable to the slider to ensure slider-clicking works
     """
-    def __init__(self, master: Page, text, default, doc, minmax=(float('-inf'), float('inf')), step=1, **kwargs):
-        super().__init__(master, text, ttk.Frame, default, doc, **kwargs)
+    def __init__(self, master: Page, key, text, default, doc, minmax=(float('-inf'), float('inf')), step=1, **kwargs):
+        super().__init__(master, key, text, ttk.Frame, default, doc, **kwargs)
         # model-binding
         self.data = self._init_data(tk.IntVar)
         # view
@@ -649,8 +839,8 @@ class FloatEntry(Entry):
     """
     - must NOT inherit from IntEntry to avoid slider malfunction
     """
-    def __init__(self, master: Page, text, default, doc, minmax, step=0.1, precision=2, **kwargs):
-        super().__init__(master, text, ttk.Frame, default, doc, **kwargs)
+    def __init__(self, master: Page, key, text, default, doc, minmax, step=0.1, precision=2, **kwargs):
+        super().__init__(master, key, text, ttk.Frame, default, doc, **kwargs)
         self.precision = precision
         # model-binding
         self.data = self._init_data(tk.DoubleVar)
@@ -694,8 +884,8 @@ class SingleOptionEntry(Entry):
     - because most clients of optionEntry use its index instead of string value, e.g., csound oscillator waveform is defined by integer among a list of options
     - we must bind to index instead of value for model-binding
     """
-    def __init__(self, master: Page, text, options, default, doc, **kwargs):
-        super().__init__(master, text, ttk.Combobox, default, doc, values=options, **kwargs)
+    def __init__(self, master: Page, key, text, options, default, doc, **kwargs):
+        super().__init__(master, key, text, ttk.Combobox, default, doc, values=options, **kwargs)
         # model-binding
         self.data = self._init_data(tk.StringVar)
         self.field.configure(textvariable=self.data, state='readonly')
@@ -728,8 +918,8 @@ class SingleOptionEntry(Entry):
 
 
 class MultiOptionEntry(Entry):
-    def __init__(self, master: Page, text, options, default, doc, **kwargs):
-        super().__init__(master, text, ttk.Menubutton, default, doc, **kwargs)
+    def __init__(self, master: Page, key, text, options, default, doc, **kwargs):
+        super().__init__(master, key, text, ttk.Menubutton, default, doc, **kwargs)
         self.data = {opt: tk.BooleanVar(name=opt, value=opt in default) for opt in options}
         self.field.configure(text='Select one ore more ...')
         # build option menu
@@ -753,7 +943,7 @@ class MultiOptionEntry(Entry):
         - serialized data: selected subset
         """
         for opt in self.data:
-            self.data[opt].set(1 if opt in values else 0)
+            self.data[opt].set(opt in values)
 
     def _select_all(self):
         for k, v in self.data.items():
@@ -778,47 +968,130 @@ class MultiOptionEntry(Entry):
 
 
 class BoolEntry(Entry):
-    def __init__(self, master: Page, text, default, doc, **kwargs):
-        super().__init__(master, text, ttk.Checkbutton, default, doc, **kwargs)
+    def __init__(self, master: Page, key, text, default, doc, **kwargs):
+        super().__init__(master, key, text, ttk.Checkbutton, default, doc, **kwargs)
         self.data = self._init_data(tk.BooleanVar)
         self.field.configure(variable=self.data)
 
 
 class TextEntry(Entry):
-    def __init__(self, master: Page, text, default, doc, **kwargs):
+    def __init__(self, master: Page, key, text, default, doc, **kwargs):
         """there is no ttk.Text"""
 
-        super().__init__(master, text, tk.Text, default, doc, height=4, **kwargs)
+        super().__init__(master, key, text, tktext.ScrolledText, default, doc, height=4, wrap=tk.WORD, **kwargs)
         self.data = self._init_data(tk.StringVar)
-        # because the binding definition below will trigger callbacks
-        # we must avoid feedback loop by setting this flag right before binding
-        self._updatingModel = True
-        self.field.bind("<<Modified>>", self._on_text_changed)
+        self.field.bind("<KeyRelease>", self._on_text_changed)
+        self.field.bind("<FocusOut>", self._on_text_changed)
         self.data.trace_add("write", self._on_data_changed)
         self.field.insert("1.0", default)
-
-    def set_data(self, value):
-        super().set_data(value)
-        self._on_data_changed()
+        # allow paste
+        self.actionBtn = ttk.Button(self, text="Paste", command=self.on_action)
+        self.actionBtn.pack(side='bottom', expand=True, padx=5, anchor="w")
 
     def _on_data_changed(self, *args):
         """
         - update view on model changes
         """
-        self._updatingModel = True
-        seen = self.field.get("1.0", tk.END).strip()
-        to_see = self.data.get()
-        if seen != to_see:
+        text_content = self.data.get()
+        if self.field.get("1.0", tk.END).strip() != text_content:
             self.field.delete("1.0", tk.END)
-            self.field.insert("1.0", to_see)
-        self._updatingModel = False
+            self.field.insert("1.0", text_content)
 
     def _on_text_changed(self, event):
         """
         - update model on user editing
         - must avoid feedback loop when text changes are caused by model changes
         """
-        if self._updatingModel:
+        self.data.set(self.field.get("1.0", tk.END).strip())
+
+    def on_action(self):
+        """
+        - replace entry text with clipboard content
+        """
+        # clear the text field first
+        self.field.delete("1.0", tk.END)
+        self.field.insert(tk.INSERT, self.field.clipboard_get())
+
+
+class FileEntry(TextEntry):
+    """
+    - user can type in a list of paths as text lines, one per line
+    - to specify a default file-extension, place it as the head of file_patterns
+    - always return a list even when there is only one; so use self.data[0] on app side for a single-file case
+    """
+    def __init__(self, master: Page, key, path, default, doc, file_patterns=(), start_dir=util.get_platform_home_dir(), **kwargs):
+        super().__init__(master, key, path, default, doc, **kwargs)
+        self.filePats = file_patterns
+        self.startDir = start_dir
+        self._fix_platform_patterns()
+        self.actionBtn.configure(text='Browse ...')
+
+    def get_data(self):
+        """
+        - a list of paths
+        """
+        return self.data.get().splitlines()
+
+    def set_data(self, value: list[str]):
+        self.data.set('\n'.join(value))
+        self._on_data_changed()
+
+    def reset(self):
+        lst = self.default if isinstance(self.default, (list, tuple)) else [self.default]
+        self.set_data(lst)
+
+    def on_action(self):
+        preferred_ext = self.filePats[pattern := 0][ext := 1]
+        selected = filedialog.askopenfilename(
+            parent=self,
+            title="Select File(s)",
+            initialdir=self.startDir,
+            filetypes=self.filePats,
+            defaultextension=preferred_ext
+        )
+        if user_cancelled := selected == '':
+            # keep current
             return
-        self.data.set(self.field.get("1.0", tk.END))
-        self.field.edit_modified(False)
+        if multi_selection := isinstance(selected, (tuple, list)):
+            selected = '\n'.join(selected)
+        self.data.set(selected)
+        # memorize last selected file's folder
+        self.startDir = osp.dirname(selected)
+
+    def _fix_platform_patterns(self):
+        """
+        - macOS demands 0 or at least 2 patterns were given if filetypes is set
+        """
+        if util.PLATFORM != 'Darwin':
+            return
+        if len(self.filePats) != 1:
+            return
+        # on macOS, only one pattern was given, so fix it
+        self.filePats = tuple([self.filePats[0], ('All Files', '*')])
+
+
+class FolderEntry(TextEntry):
+    """
+    - tkinter supports single-folder selection only
+    - multiple folders can be pasted into the text field
+    """
+    def __init__(self, master: Page, key, path, default, doc, start_dir=util.get_platform_home_dir(), **kwargs):
+        super().__init__(master, key, path, default, doc, **kwargs)
+        self.startDir = start_dir
+        self.actionBtn.configure(text='Browse ...')
+
+    def get_data(self):
+        return self.data.get().splitlines()
+
+    def on_action(self):
+        selected = filedialog.askdirectory(
+            parent=self,
+            title="Select Folder(s)",
+            initialdir=self.startDir,
+        )
+        if user_cancelled := selected == '':
+            # keep current
+            return
+        self.data.set(selected)
+        # memorize last selected file's folder
+        self.startDir = osp.dirname(selected)
